#ifndef bqtRGBhh
#define bqtRGBhh

#include "floatdefs.hh"
#include <algorithm>

using std::max;

class RGB
{
protected:
    FLOAT r,g,b;
private:
    inline const int intensity(FLOAT elem) const
    {
    #if 0
        if(elem < 0)return 0;
        if(elem <= 1)return int(255*elem);
        if(elem > 1)return 255;
    #else
        if(elem < 0)return 0;
        const FLOAT maxi = max(max(r,g), b);
        if(maxi <= 1)return int(255*elem);
        if(elem > 1)return 255;
        const FLOAT jaettavaa = maxi-1;
        int v = int((elem + jaettavaa)*255);
        return v>=255 ? 255 : v;
    #endif
    }

public:
    RGB() : r(0), g(0), b(0)
    {
    }
    RGB(FLOAT_ARG R, FLOAT_ARG G, FLOAT_ARG B) : r(R), g(G), b(B)
    {
    }
    
    void Set(FLOAT_ARG R, FLOAT_ARG G, FLOAT_ARG B) { r=R; g=G; b=B; }

    /* Reflect by pigment */
    const RGB &operator *=(const RGB &p) { r*=p.r; g*=p.g; b*=p.b; return *this; }
    /* Add light */
    const RGB &operator +=(const RGB &p) { r+=p.r; g+=p.g; b+=p.b; return *this; }
    /* ? */
    const RGB &operator -=(const RGB &p) { r-=p.r; g-=p.g; b-=p.b; return *this; }
    /* Brighten */
    const RGB &operator *=(FLOAT_ARG p) { r*=p; g*=p; b*=b; return *this; }
    /* Darken */
    const RGB &operator /=(FLOAT_ARG p) { r/=p; g/=p; b/=b; return *this; }
    const RGB operator *(const RGB &p) const { return RGB(r*p.r, g*p.g, b*p.b); }
    const RGB operator +(const RGB &p) const { return RGB(r+p.r, g+p.g, b+p.b); }
    const RGB operator -(const RGB &p) const { return RGB(r-p.r, g-p.g, b-p.b); }
    const RGB operator +(FLOAT_ARG p) const { return RGB(r+p, g+p, b+p); }
    const RGB operator -(FLOAT_ARG p) const { return RGB(r-p, g-p, b-p); }
    const RGB operator *(FLOAT_ARG p) const { return RGB(r*p, g*p, b*p); }
    const RGB operator /(FLOAT_ARG p) const { return RGB(r/p, g/p, b/p); }

    const unsigned char getR() const { return intensity(r); } /* 0..255 red */
    const unsigned char getG() const { return intensity(g); } /* 0..255 green */
    const unsigned char getB() const { return intensity(b); } /* 0..255 blue */

    const FLOAT power() const { return r+g+b; }
    
    const FLOAT luminance() const
    {
        FLOAT Y = 0.299 * r + 0.587 * g + 0.114 * b;
        //FLOAT Y = 0.212671 * r + 0.715160 * g + 0.072169* b;
        return Y;
    };

    void print() const
    {
        printf("<%.2"PF" %.2"PF" %.2"PF">", r,g,b);
    }
};

class Pigment : public RGB
{
public:
    /* Diffuse light multiplier.
     * Defines the coarsiness of the surface.
     * Range: 0..1
     * 1 means that it reflects light sources not depending of directions.
     */
    FLOAT diffmul;
    
    /* Works like diffmul, but ignores all blocking stuff. */
    FLOAT dirdiffmul;
    
    /* Mirrored light multiplier. */
    /* Range: 0..1 */
    FLOAT specmul;
    
    /* How much of the behind stuff shows. 1=invisible wall. */
    FLOAT penetration;
public:
    Pigment(FLOAT_ARG R, FLOAT_ARG G, FLOAT_ARG B) : RGB(R,G,B) { setdefault(); }
    Pigment() { setdefault(); }
    void setdefault()
    {
        diffmul     = 0.7;
        dirdiffmul  = 0.0;
        specmul     = 0.9;
        penetration = 0.0;
    }
};

#endif
